/*
 * Download driver for excelan ethernet driver
 *
 * This driver is used to:
 *	- download software to run in the board
 *
 * Written by: Kipp Hickman
 * Inspired by: Bill Northlich
 */

#ifdef SYSTEM5
#include "sys/param.h"
#include "sys/config.h"
#include "sys/types.h"
#include "sys/systm.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/tty.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/buf.h"
#endif

#ifdef VERSION7
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/proc.h>
#include <signal.h>
#endif
#ifdef PEGASUS
#include <sys/mmu.h>
#endif

#include <sys/extypes.h>
#include <sys/socket.h>
#include <sys/swap.h>
#include <sys/init.h>
#include <sys/exos.h>
#include <sys/exreg.h>
#include <ex_ioctl.h>

#define MIN(a,b)	(((a) < (b))? (a):(b))
struct	buf *ex_getbuf();
struct	msg *ex_findmsg();
struct	kludge *ex_send();
extern	struct msg_area *msgarea;
extern	struct exctrl ex_db;
#ifdef PEGASUS
#define vtop(x)		((long)x + (long)dscraddr(SEGKD))
#endif

/* variables unique to the downloader */
static	long dladdr;			/* down load address, in exos memory */
static	char dlbusy;			/* busy flag */
static	struct buf *dlbuf;		/* pointer to buf containing memory */

#undef	DEBUG
#undef	DEBUG_NOISE
#undef	DEBUG_QUEUES
#define bswap(x) (((x)<<8)|(((x)>>8)&0xff))

/*
 * dlopen:
 *	- init down load stuff
 *	- blast any procs we might have hanging on the socket
 *		- this way, unix procs get reset when board gets reset
 *	- steal a buffer from buffer cache for use during down load
 */
/*ARGSUSED*/
dlopen(dev, flag)
	dev_t dev;
	int flag;
{
	register struct kludge *kp;
#ifdef	DEBUG_NOISE
printf("dlopen(%x, %d)\n", dev, flag);
#endif
	if (!suser())
		return;
	if (dlbusy)
		u.u_error = EBUSY;
	else {
		for (kp = kludge; kp < &kludge[NET_KLUDGES]; kp++) {
			if (kp->kl_state & KL_WAITING)
				psignal(kp->kl_proc, SIGKILL);
			kp->kl_state = KL_FREE;
		}
#ifdef PEGASUS
		if (ex_db.ex_ini == 0)
#else
		if (ex_db.ex_init == 0)
#endif
			exinit();
		ex_db.ex_netdown = 1;
		dladdr = 0;
		dlbusy = 1;
		dlbuf = ex_getbuf();
		if (exsetup(EXOS_HOSTLOAD))
			u.u_error = EIO;
	}
}

/*
 * dlclose:
 *	- clean up after down loading
 */
/*ARGSUSED*/
dlclose(dev, flag)
	dev_t dev;
	int flag;
{
#ifdef	DEBUG_NOISE
printf("dlclose(%x, %d)\n", dev, flag);
#endif
	dlbusy = 0;
	brelse(dlbuf);
}

/*
 * dlwrite:
 *	- down load driver routine for exos
 */
/*ARGSUSED*/
dlwrite(dev)
	dev_t dev;
{
	register struct msg *mp;
	register struct net_msg *np;
	register struct net_dload *net_dload;
	register struct exos_paddr *ap;
	register struct kludge *kp;
	register caddr_t baddr = dlbuf->b_un.b_addr;
	register int amount;

#ifdef	DEBUG
printf("dlwrite(%x): u.u_count=%d\n", dev, u.u_count);
#endif
	while (u.u_count) {
		amount = MIN(BSIZE, u.u_count);
		iomove(baddr, amount, B_WRITE);
		if (u.u_error)
			break;
#ifdef DEBUG_QUEUES
{
int i;
short *p = (short *)baddr;

for (i = 0; i < 16; i++)
printf("%x ", *p++);
printf("\n");
}
#endif
#ifdef	DEBUG_NOISE
/*printf("dlwrite: Sending DLOAD message\n");*/
#endif

	    /* get message buffer; send download message */

		mp = ex_findmsg();
		np = &mp->msg_msg;
		net_dload = &np->nm_u.nm_dload;
		net_dload->nm_length = amount;
#ifdef PEGASUS
		net_dload->nm_source = (long)vtop(baddr);
#else
		ap = (struct exos_paddr *)(&net_dload->nm_source);
		SET_PADDR(ap, (int)baddr);
#endif
		ap = &net_dload->nm_dest;
		SET_ADDR(ap, dladdr);
#ifdef DEBUG_QUEUES
{
int i;
short *p = (short *)net_dload;
long sel;
unsigned short lim;

for (i = 0; i < (sizeof (struct net_dload))/2 ; i++)
printf("%x ", *p++);
printf("\n");

sel = dscraddr(SEGKD);
lim = dscrlimit(SEGKD);
printf("DLWRITE:  addr(SEGKD)=%x%x limit(SEGKD)=%x b_xmem=%x\n",
(int)((sel>>16)&0xffff),(int)(sel&0xffff),lim,dlbuf->b_xmem);
}
#endif
		kp = ex_send(NET_DLOAD, mp, np, 0);
		kp->kl_state = 0;
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
#ifdef	DEBUG_NOISE
printf("dlwrite: DLOAD message sent (%x,%x)\n",ap->pa_segment,ap->pa_offset);
#endif
		if (kp->kl_reply) {
			u.u_error = EIO;
			break;
		}
		dladdr += amount;
	}
}

/*
 * dlioctl:
 *	- process down loader ioctl's
 */
/*ARGSUSED*/
dlioctl(dev, cmd, addr, flag)
	dev_t dev;
	int cmd, flag;
	caddr_t addr;
{
	register struct msg *mp;
	register struct net_msg *np;
	register struct exos_paddr *ap;
	register struct kludge *kp;
	long start;

	switch (cmd) {
	  case EXIOCSTRT:		/* start program running */
		if (copyin(addr, (caddr_t)&start, sizeof(start))) {
			u.u_error = EFAULT;
			break;
		}
#ifdef	DEBUG_NOISE
printf("dlioctl: start address=%x\n", start);
#endif
		mp = ex_findmsg();
		np = &mp->msg_msg;
		ap = &np->nm_u.nm_begin.nm_start;
		SET_ADDR(ap, start);
		kp = ex_send(NET_START, mp, np, 0);
#ifdef	DEBUG_NOISE
printf("dlioctl: START message sent (%x,%x)\n",ap->pa_segment,ap->pa_offset);
#endif
		if (kp->kl_reply)
			u.u_error = EIO;
		break;
	  case EXIOCDLOAD:		/* set dladdr */
		if (copyin(addr, (caddr_t)&dladdr, sizeof(dladdr))) {
			u.u_error = EFAULT;
			break;
		}
#ifdef	DEBUG_NOISE
printf("dlioctl: dladdr=%x\n", dladdr);
#endif
		break;
	}
}
